/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : create.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Creates a timer instance, optionally storing data from   */
/*                   the current message for inclusion in any timer message(s)*/
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: July 2001                      */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "uuidgen.h"      /* Specific header for this dll                      */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "uuid.h"

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

/* local function prototypes */
int copyUuidValueToElement(IN CciElement *elementP, 
						   IN uuid_t *uuidP,
						   IN bool_t newElement,
						   IN bool_t xmlMessage,
						   IN bool_t includeSeparators);

/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* insertAttributeTableEntry                                                  */
/* insertInputTerminalListEntry                                               */
/* insertOutputTerminalListEntry                                              */
/* getInputTerminalHandle                                                     */
/* getOutputTerminalHandle                                                    */
/* traverseMessage                                                                           */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* evaluateUuidCreate                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context         (input) - pointer to the current context                  */
/*  destinationList (input) - address of the input destination list for the   */
/*                            message                                         */
/*  exceptionList   (input) - address of the exception list for the message   */
/*  message         (input) - address of the input message object             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  None.                                                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Creates a Uuid value and stores it in the current message at the location  */
/* specified by the CopyValueTo attribute.                                    */
/*                                                                            */
/******************************************************************************/
int evaluateUuidCreate(context_t* contextP,
 			   		   CciMessage *destinationList,
		               CciMessage *exceptionList,
		               CciMessage *message)
{
	int rc,
		retCode,
		messageNumber;
	uuid_t uuid;
	CciMessageContext *msgContextP;
	CciMessage *newMsg;
	CciElement *copyToElement;
	char mbString[MAX_STRING_SIZE + 1],
		 msgBuffer[1000],
		 nodeNameMB[100];
	deployedUuidgen_t *deployedUuidgenP;
	bool_t newElementCreated,
		   xmlMessage;
	CciChar elementPath[10 * MAX_STRING_SIZE + 1],
		    *elementPathP,
			*dotCharP;
    static char *functionName="evaluateUuidCreate";

    TRACE_ENTRY("evaluateUuidCreate", ("contextP=%.8x, destinationList=%.8x, "
		                     "exceptionList=%.8x, message=%.8x",
							 contextP, destinationList,
							 exceptionList, message));

	deployedUuidgenP = (deployedUuidgen_t *)(contextP->nodeDataP);

	/* create a new UUID value */
	threadMutexLock(uuidGenMutex);
	rc = uuid_create(&uuid);
	threadMutexUnlock(uuidGenMutex);

	if (rc != UUID_SUCCESS)
	{
		if (rc == UUID_FAILED_TO_GET_MACHINE_HARDWARE_ADDRESS)
		{
   	        TRACE_RETURN(RSN_020, TT_FAILED_TO_GET_MACHINE_HARDWARE_ADDRESS);
		}
        TRACE_RETURN(RSN_021, UUID_FAILED);
	}

	TRACE_MID(TRACE_INFO, ("Generated UUID value is..."));
	TRACE_MIDHEX(TRACE_INFO, &uuid, sizeof(uuid));

	/* create a copy of message so we can insert the uuid value */
	msgContextP = cniGetMessageContext(&rc, message);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_014, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

    newMsg = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new message "
			                    "(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_015, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}
	TRACE_MID(TRACE_INFO, ("New message created at %.8x", newMsg));

	rc = copyMessage(newMsg, message);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old message to new message"));
		TRACE_RETURN(RSN_034, TT_FAILED_TO_CREATE_NEW_MESSAGE);
	}

	elementPathP = deployedUuidgenP->copyValueToLocation;
	xmlMessage = messageBodyIsXML(newMsg);
	if (xmlMessage)
	{
		/* 
		 * If path does not specify a known child element of root, then modify it if it 
		 * doesn't begin with Root.XML.
		 */
		if (UCncmp(C.ROOT_DOT_MQMD_DOT, elementPathP, UClen(C.ROOT_DOT_MQMD_DOT)) &&
		    UCncmp(C.ROOT_DOT_PROPERTIES_DOT, elementPathP, UClen(C.ROOT_DOT_PROPERTIES_DOT)))
		{
			TRACE_MID(TRACE_INFO, ("Checking path '%S' for XML qualifier", elementPathP));
			/* modify specified path if it doesn't begin with Root.XML. */
			if (UCncmp(C.ROOT_DOT_XML_DOT, elementPathP, UClen(C.ROOT_DOT_XML_DOT)))
			{
				/* specified path doesn't include XML qualifier, so modify it */
				TRACE_MID(TRACE_INFO, ("Adding XML qualifier to path"));

				/* build modified path string */
				UCcpy(elementPath, C.ROOT_DOT_XML_DOT);

				dotCharP = UCchr(elementPathP, C.DOT_CHAR);
				dotCharP++;  // move to next character following dot


				UCcat(elementPath, dotCharP);
				TRACE_MID(TRACE_INFO, ("Modified path is '%S'", elementPath));

				elementPathP = elementPath;
			}
		}
	}

	/* locate (and create if necessary) specified output element */
	copyToElement = locateElementInMessage(newMsg,
		                                   elementPathP,
									       TRUE,  /* create new element(s) if required */  
									       NULL,
									       &newElementCreated);
	/* 
	 * Propagate message to the 'failure' terminal if we can't find/create the element
	 */
    if (!copyToElement)
	{

		UCsprintf(msgBuffer, "Failed to locate specified syntax element '%S' in message",
							  deployedUuidgenP->copyValueToLocation);

        cciLog(&rc, 
		       CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   MJB_FAILED_TO_LOCATE_SYNTAX_ELEMENT, 
			   msgBuffer,
#if !defined(WIN32)
		       brokerName,
#endif
			   ATTR_COPY_VALUE_TO_MB,
 			   /* convert copyToLocation to multi byte string */
			   UCToStr(deployedUuidgenP->copyValueToLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		/* delete the new message */
		cniDeleteMessage(&rc,
						 newMsg);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
			                   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* copy the uuid to the located (or created) element */
	retCode = copyUuidValueToElement(copyToElement, 
		                             &uuid, 
									 newElementCreated,
									 xmlMessage,
									 deployedUuidgenP->includeSeparators);
	if (retCode != TT_SUCCESS)
	{
		/* handle a variety of errors here... */
		if (retCode == TT_FAILED_TO_SET_ELEMENT_VALUE)
		{
			UCsprintf(msgBuffer, "Failed to store UUID value in message at syntax element '%S'",
								deployedUuidgenP->copyValueToLocation);
			messageNumber = MJB_FAILED_TO_STORE_UUID_VALUE_IN_SYNTAX_ELEMENT;
		}
		else
		if (retCode == TT_INVALID_ELEMENT_TYPE_SPECIFIED)
		{
			UCsprintf(msgBuffer, "The specified element for storing the UUID value is not "
				                 "suitable for this operation.  The syntax element is '%S'",
								deployedUuidgenP->copyValueToLocation);
			messageNumber = MJB_INCOMPATIBLE_SYNTAX_ELEMENT_SPECIFIED_FOR_UUID;
		}
		else
		{
			UCsprintf(msgBuffer, "Unknown error occurred when storing the UUID value at the "
				                 "specified syntax element '%S'",
								deployedUuidgenP->copyValueToLocation);
			messageNumber = MJB_UNKNOWN_ERROR_WHEN_STORING_UUID;
		}

		UCsprintf(nodeNameMB, "'%S'", contextP->nodeName);

		cciLog(&rc, 
			   CCI_LOG_ERROR, 
			   __FILE__, 
			   __LINE__, 
			   functionName, 
			   C.PLUGIN_MESSAGE_CATALOGUE, 
			   messageNumber, 
			   msgBuffer,
#if !defined(WIN32)
			   brokerName,
#endif
			   ATTR_COPY_VALUE_TO_MB,
			   nodeNameMB,
 			   /* convert copyToLocation to multi byte string */
			   UCToStr(deployedUuidgenP->copyValueToLocation,
					   mbString,
					   sizeof(mbString),
					   NULL),
			   (char *)0);

		/* delete the new message */
		cniDeleteMessage(&rc,
						 newMsg);
		
		TRACE_MID(TRACE_INFO, ("About to propagate message at %.8x to 'failure' "
							   "terminal", message));
		retCode = propagateMessage(contextP,
								   C.TERM_FAILURE,
								   message,
								   destinationList,
								   exceptionList,
								   FALSE); /* this is not a new message */

		if (retCode != TT_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'failure' terminal"));
			RETHROW_IF_EXCEPTION(retCode);
			TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
		}

		TRACE_RETURN(RSN_013, TT_SUCCESS);
	}

	/* propagate message */
	TRACE_MID(TRACE_INFO, ("Propagating message to 'out' terminal"));
	retCode = propagateMessage(contextP,
							   C.TERM_OUT,
							   newMsg,
							   destinationList,
							   exceptionList,
						       TRUE); /* this is a new message */

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to propagate message to 'out' terminal"));
		RETHROW_IF_EXCEPTION(retCode);
		TRACE_RETURN(RSN_012, TT_PROPAGATION_ERROR);
	}

	/* delete the new message */
    cniDeleteMessage(&rc,
		             newMsg);

	TRACE_RETURN(RSN_019, TT_SUCCESS);
}


/* local functions */
#define DEFAULT_UUID_BUFFER_SIZE 10000
/******************************************************************************/
/* copyUuidValueToElement                                                     */
/******************************************************************************/
int copyUuidValueToElement(IN CciElement *elementP, 
						   IN uuid_t *uuidP,
						   IN bool_t newElement,
						   IN bool_t xmlMessage,
						   IN bool_t includeSeparators)
{
    struct CciByteArray byteArrayDetails,
		                tempByteArrayDetails;
    struct CciBitArray tempBitArrayDetails,
                       bitArrayDetails;
	int rc,
		numBytes,
		numCciChars;
	bool_t valueAlreadyExists = FALSE,
		   compatibleDataType,
		   storeAsHex = FALSE;
	CciElementType elementType,
		           childElementType;
	CciElement *childElementP,
		       *dataElementP;
	CciValueType valueType;
	char uuidBuffer[DEFAULT_UUID_BUFFER_SIZE],
		 *uuidBufferP = uuidBuffer,
		 hexUuidString[(sizeof(uuid_t) + 4 + 1) * 2], /* allow for 4 separator characters */
		 tempHexUuidString[sizeof(hexUuidString)];
	CciChar smallBuffer[1],
		    hexUuidStringUC[sizeof(hexUuidString)];
    TRACE_ENTRY("copyUuidValueToElement", ("elementP=%.8x, uuidP=%.8x, newElement=%s, "
		                                   "includeSeparators=%s",
							                elementP, uuidP,
											newElement ? "TRUE" : "FALSE",
											xmlMessage ? "TRUE" : "FALSE",
											includeSeparators ? "TRUE" : "FALSE"));

	byteArrayDetails.pointer = uuidP;
	byteArrayDetails.size = sizeof(uuid_t);

	dataElementP = elementP;

	/* check target element is suitable to store the UUID value */
	elementType = cniElementType(&rc, 
								 elementP);
	TRACE_MID(TRACE_INFO, ("elementType is %s",
		elementType == CCI_ELEMENT_TYPE_NAME ? "CCI_ELEMENT_TYPE_NAME" :
		elementType == CCI_ELEMENT_TYPE_NAME_VALUE ? "CCI_ELEMENT_TYPE_NAME_VALUE" :
		elementType == CCI_ELEMENT_TYPE_VALUE ? "CCI_ELEMENT_TYPE_VALUE" : "UNKNOWN"));

	if (elementType != CCI_ELEMENT_TYPE_NAME &&
	    elementType != CCI_ELEMENT_TYPE_NAME_VALUE &&
	    elementType != CCI_ELEMENT_TYPE_VALUE)
	{
		TRACE_MID(TRACE_ERROR, ("Element type is not suitable for storing UUID value"));
		TRACE_RETURN(RSN_019, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
	}

	/* if its a NAME type element, then address its value element, or create one */
	if (elementType == CCI_ELEMENT_TYPE_NAME)
	{
		/* address the value element if it exists */
		childElementP = cniFirstChild(&rc,
                                      elementP);
		if (childElementP)
		{
			TRACE_MID(TRACE_INFO, ("Child element exists at %.8x", childElementP));
			/* is it a NameValue type? */
			childElementType = cniElementType(&rc, 
											  childElementP);
			if (childElementType != CCI_ELEMENT_TYPE_VALUE)
			{
				/* 
				 * At this stage, we could delete everything under the NAME element and create
				 * a new VALUE element, but for the moment we'll fail this attempt.
				 */
				TRACE_MID(TRACE_ERROR, ("Element type is not suitable for storing UUID value"));
				TRACE_RETURN(RSN_019, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
			}
			valueType = cniElementValueType(&rc, childElementP);
			TRACE_MID(TRACE_INFO, ("valueType is %d", valueType));
			valueAlreadyExists = TRUE;
		}
		else
		{
			TRACE_MID(TRACE_INFO, ("Child element being created"));
			// create a child VALUE element
			childElementP = createChildElement(elementP,
				                               NULL,
											   CCI_ELEMENT_TYPE_VALUE);
			TRACE_MID(TRACE_INFO, ("Child element created at %.8x", childElementP));
		}
		dataElementP = childElementP;
	}

	/* convert uuid value to hex chars */
	if (xmlMessage || newElement || valueAlreadyExists && valueType == CCI_VALUE_TYPE_CHAR)
	{
		storeAsHex = TRUE;
		dispHex((char *)uuidP,
				sizeof(uuid_t),
				hexUuidString);

		/* add separators if requested */
		if (includeSeparators)
		{
			// e.g. turn 36e3f290d91f11d5a3e6204c4f4f5020 into
			//            36e3f290-d91f-11d5-a3e6-204c4f4f5020
			TRACE_MID(TRACE_INFO, ("Adding hyphens as separators"));
			strcpy(tempHexUuidString, hexUuidString);
			hexUuidString[8] = '\0';
			strcat(hexUuidString, "-");
			strncat(hexUuidString,tempHexUuidString+8, 4);
			strcat(hexUuidString, "-");
			strncat(hexUuidString,tempHexUuidString+12, 4);
			strcat(hexUuidString, "-");
			strncat(hexUuidString,tempHexUuidString+16, 4);
			strcat(hexUuidString, "-");
			strcat(hexUuidString,tempHexUuidString+20);
			TRACE_MID(TRACE_INFO, ("Uuid: %s (with separators: %s",
					               tempHexUuidString, hexUuidString));
		}

		/* convert multibyte string to Unicode */
        strToUC(hexUuidString,
				strlen(hexUuidString),
				hexUuidStringUC,
				NULL);

		if (xmlMessage || newElement)
		{
			TRACE_MID(TRACE_INFO, ("setting CHAR value into message"));
			numCciChars = UClen(hexUuidStringUC);
			cniSetElementCharacterValue(&rc, 
										dataElementP, 
										hexUuidStringUC,
										numCciChars);
			if (rc)
			{
				TRACE_MID(TRACE_ERROR, ("Failed to set UUID value into specified element"));
				TRACE_RETURN(RSN_019, TT_FAILED_TO_SET_ELEMENT_VALUE);
			}
			TRACE_MID(TRACE_INFO, ("UUID value successfully assigned to new element"));
			TRACE_RETURN(RSN_019, TT_SUCCESS);
		}
	}

    /* 
	 * For elements that already exist, we do not want to change the length of the 
	 * value field.
	 * If a value already exists, then ensure that we replace it with a value whose length
	 * is the same - pad with binary zeroes (or blanks) if necessary
	 */
	compatibleDataType = FALSE;
	switch (valueType)
	{
		/************************************************/ 
		/* Check for compatible data types              */
		/************************************************/ 
		case CCI_VALUE_TYPE_BYTEARRAY:
			TRACE_MID(TRACE_INFO, ("element value type is BYTEARRAY"));
			tempByteArrayDetails.pointer = (void *)smallBuffer;
			tempByteArrayDetails.size = sizeof(smallBuffer);
			numBytes = cniElementByteArrayValue(&rc,
												dataElementP,
												&tempByteArrayDetails);
			compatibleDataType = TRUE;
			break;

		case CCI_VALUE_TYPE_BITARRAY:
			TRACE_MID(TRACE_INFO, ("element value type is BITARRAY"));
			tempBitArrayDetails.pointer = (void *)smallBuffer;
			tempBitArrayDetails.size = sizeof(smallBuffer);
			numBytes = cniElementBitArrayValue(&rc,
											   dataElementP,
											   &tempBitArrayDetails);
			compatibleDataType = TRUE;
			break;

		case CCI_VALUE_TYPE_DECIMAL:
			TRACE_MID(TRACE_INFO, ("element value type is DECIMAL"));
			break;

		case CCI_VALUE_TYPE_CHAR:
			TRACE_MID(TRACE_INFO, ("element value type is CHAR"));
			numBytes = cniElementCharacterValue(&rc, 
										 		dataElementP, 
												smallBuffer,
						                        sizeof(smallBuffer));
			numBytes *= sizeof(CciChar);
			compatibleDataType = TRUE;
			break;

		/*********************************************/ 
		/* The following are fixed length data types */
		/*********************************************/ 
		case CCI_VALUE_TYPE_BOOL:
			TRACE_MID(TRACE_INFO, ("element value type is BOOL"));
			break;

		case CCI_VALUE_TYPE_INTEGER:
			TRACE_MID(TRACE_INFO, ("element value type is INTEGER"));
			break;

		case CCI_VALUE_TYPE_REAL:
			TRACE_MID(TRACE_INFO, ("element value type is REAL"));
			break;

		case CCI_VALUE_TYPE_TIME:
			TRACE_MID(TRACE_INFO, ("element value type is TIME"));
			break;

		case CCI_VALUE_TYPE_GMTTIME:
			TRACE_MID(TRACE_INFO, ("element value type is GMTTIME"));
			break;

		case CCI_VALUE_TYPE_DATE:
			TRACE_MID(TRACE_INFO, ("element value type is DATE"));
			break;

		case CCI_VALUE_TYPE_TIMESTAMP:
			TRACE_MID(TRACE_INFO, ("element value type is TIMESTAMP"));
			break;

		case CCI_VALUE_TYPE_GMTTIMESTAMP:
			TRACE_MID(TRACE_INFO, ("element value type is GMTTIMESTAMP"));
			break;

		default:
			TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
			// return error
			break;
	}

	if (!compatibleDataType)
	{
		TRACE_MID(TRACE_ERROR, ("Element type is not suitable for storing UUID value"));
		TRACE_RETURN(RSN_019, TT_INVALID_ELEMENT_TYPE_SPECIFIED);
	}

	/* make sure we don't change the length of the existing data value */
	if (numBytes > DEFAULT_UUID_BUFFER_SIZE)
	{
		uuidBufferP = (char *)malloc(numBytes);
		TRACE_MID(TRACE_INFO, ("%d bytes allocated for uuidBufferP at %.8x", 
			                   numBytes, uuidBufferP));
	}

	if (storeAsHex)
	{
		/* clear buffer to blanks */
	    memset(uuidBufferP, ' ', numBytes);
		memcpy(uuidBufferP, hexUuidStringUC, 
			   numBytes < (int)(UClen(hexUuidStringUC) * sizeof(CciChar)) ? numBytes : 
		                               (UClen(hexUuidStringUC) * sizeof(CciChar)));
	}
	else
	{
		/* clear buffer to binary zeroes */
	    memset(uuidBufferP, 0, numBytes);
		memcpy(uuidBufferP, uuidP, numBytes < sizeof(uuid_t) ? numBytes : sizeof(uuid_t));
	}

	/* set the value into the element */
	switch (valueType)
	{
		/************************************************/ 
		/* The following are variable length data types */
		/************************************************/ 
		case CCI_VALUE_TYPE_BYTEARRAY:
			byteArrayDetails.pointer = uuidBufferP;
			byteArrayDetails.size = numBytes;
			cniSetElementByteArrayValue(&rc,
										dataElementP,
										&byteArrayDetails);
			break;

		case CCI_VALUE_TYPE_BITARRAY:
			bitArrayDetails.pointer = uuidBufferP;
			bitArrayDetails.size = numBytes;
			cniSetElementBitArrayValue(&rc,
									   dataElementP,
									   &bitArrayDetails);
			break;

		case CCI_VALUE_TYPE_CHAR:
			numCciChars = UClen((CciChar *)uuidBufferP);
			cniSetElementCharacterValue(&rc, 
										dataElementP, 
										(CciChar *)uuidBufferP,
										numCciChars);
			break;

		default:
			TRACE_MID(TRACE_INFO, ("element value type is (default) UNKNOWN"));
			break;
	}
	if (rc)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set UUID value into specified element"));
		TRACE_RETURN(RSN_019, TT_FAILED_TO_SET_ELEMENT_VALUE);
	}
	TRACE_MID(TRACE_INFO, ("UUID value successfully assigned to element"));

	/* free new buffer if allocated */
	if (uuidBufferP != uuidBuffer)
	{
		free(uuidBufferP);
		TRACE_MID(TRACE_INFO, ("uuidBufferP free'd"));
	}

	TRACE_RETURN(RSN_019, TT_SUCCESS);
}
