#define DECLARE_PLUGIN_DATA 1
/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : plugin.c                                                 */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : These are common plugin node utility functions.          */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Apr 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "plugin.h"		/* Function prototypes for common plugin node funcs   */


/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* insertAttributeTableEntry                                                  */
/* insertInputTerminalListEntry                                               */
/* insertOutputTerminalListEntry                                              */
/* getInputTerminalHandle                                                     */
/* getOutputTerminalHandle                                                    */
/* traverseMessage                                                            */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* insertAttributeTableEntry                                                  */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input) - pointer to the current context                    */
/*  nodeName      (input) - node name                                         */
/*  attributeName (input) - attribute name to be inserted into table          */
/*  type          (input) - type of the attribute                             */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to newly created attribute table entry                            */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for attribute and link it to the current context.     */
/******************************************************************************/
attributeTableEntry_t *insertAttributeTableEntry(context_t *contextP,
                                                 CciChar *nodeName,
												 bool_t customAttribute,
                                                 CciChar *attributeName,
												 int type)
{
    attributeTableEntry_t *entryP;
    int attributeSize;
    TRACE_ENTRY("insertAttributeTableEntry", ("contextP=%.8x, nodeName=%S, "
											  "attributeName=%S, type=%d",
											   contextP, nodeName,
											   attributeName, type));

    /* Allocate storage for the attribute entry */
    entryP = (attributeTableEntry_t *)malloc(sizeof(attributeTableEntry_t));

    /* if malloc is unsuccessful, return null */
    if (!entryP)
	{
		/* gen error message */
		TRACE_RETURN_TYPE(RSN_001, attributeTableEntry_t *, entryP);
	}
	TRACE_ALLOC("plugin.c-31", entryP);

	/* initialise to binary zeroes */
	memset(entryP, 0, sizeof(attributeTableEntry_t));

	/* initialise the entry */
	entryP->type = type;

	/* custom attribute entry? */
	entryP->customAttribute = customAttribute;

    /* set the attribute size depending on the type of attribute */
	switch (entryP->type)
	{
	    case CCI_VALUE_TYPE_INTEGER:
	    attributeSize = MAX_STRING_SIZE;
		break;

	    case CCI_VALUE_TYPE_BOOL:
		    attributeSize = sizeof(bool_t);
		    break;
  
	    case CCI_VALUE_TYPE_BYTEARRAY:
		    attributeSize = MAX_STRING_SIZE;
		    break;
  
        default:
		    break;
	}

	/* acquire and clear storage for attribute value */
	entryP->valuePtr = (void *)malloc(attributeSize);

	if (!entryP->valuePtr)
	{
		free(entryP);
		TRACE_FREE("plugin.c-31", entryP);
		TRACE_RETURN_TYPE(RSN_002, attributeTableEntry_t *, NULL);
	}
	TRACE_ALLOC("plugin.c-32", entryP->valuePtr);

	/* store away the attribute name */
	UCcpy (entryP->name, attributeName);

	/* initialise the attribute value storage */
    memset(entryP->valuePtr, 0, attributeSize);

	/* add this entry to the head of the attribute entry chain */
	entryP->nextP = contextP->attributeListHead;
	contextP->attributeListHead = entryP;

	/* return new entry to caller */
	TRACE_RETURN_TYPE(RSN_003, attributeTableEntry_t *, entryP);
}

/******************************************************************************/
/* Deletes the attribute storage linked to the current context.               */
/******************************************************************************/
void deleteAttributeTableEntries(context_t *contextP)
{
    attributeTableEntry_t *entryP,
		                  *nextEntryP;
    TRACE_ENTRY("deleteAttributeTableEntries", ("contextP=%.8x", contextP));

	entryP = contextP->attributeListHead;

	while(entryP)
	{
		TRACE_MID(TRACE_INFO, ("processing table entry (%S) at %.8x",
			                   entryP->name, entryP));
		nextEntryP = entryP->nextP;

		/* free the value if set */
		if (entryP->valuePtr)
		{
			TRACE_MID(TRACE_INFO, ("Releasing storage for entry value at %.8x",
								   entryP->valuePtr));
			TRACE_FREE("plugin.c-32", entryP->valuePtr);
			free(entryP->valuePtr);
		}

		/* free the entry itself */
		TRACE_MID(TRACE_INFO, ("Releasing storage for table entry at %.8x",
							   entryP));
        free(entryP);
		TRACE_FREE("plugin.c-31", entryP);

		entryP = nextEntryP;
	}

	/* return to caller */
	TRACE_RETURN_VOID(RSN_004);
}

/******************************************************************************/
/* insertInputTerminalListEntry                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  Pointer to an input terminal list entry.                                  */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for input terminal entry and link it to the current   */
/* context.                                                                   */
/******************************************************************************/
terminalListEntry_t *insertInputTerminalListEntry(context_t *contextP,
											      CciChar *terminalName)
{
    terminalListEntry_t *entryP;
    int rc;
    TRACE_ENTRY("insertInputTerminalListEntry", ("contextP=%.8x, terminalName=%S",
											      contextP, terminalName));

    entryP = (terminalListEntry_t *)malloc(sizeof(terminalListEntry_t));
    if (!entryP)
	{
		/* gen error message */
	    TRACE_RETURN_TYPE(RSN_005, terminalListEntry_t *, entryP);
	}
	TRACE_ALLOC("plugin.c-33", entryP);

	/* initialise the terminal entry */

    /* Store the terminal name */
    UCcpy(entryP->name, terminalName);

    /* Create terminal and save its handle */
    entryP->handle = cniCreateInputTerminal(&rc, 
		                                    contextP->nodeObject, 
											terminalName);

	/* add this entry to the head of the terminal entry chain */
	entryP->next = contextP->inputTerminalListHead;
	contextP->inputTerminalListHead = entryP;

    TRACE_RETURN_TYPE(RSN_006, terminalListEntry_t *, entryP);
}

/******************************************************************************/
/* insertOutputTerminalListEntry                                              */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal to be inserted into list          */
/*                                                                            */
/*  Pointer to an output terminal list entry.                                 */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Acquires the storage for output terminal entry and link it to the current  */
/* context.                                                                   */
/******************************************************************************/
terminalListEntry_t *insertOutputTerminalListEntry(context_t *contextP,
								   			       CciChar *terminalName)
{
    terminalListEntry_t *entryP;
    int rc;
    TRACE_ENTRY("insertOutputTerminalListEntry", ("contextP=%.8x, "
                                                  "terminalName=%S",
											       contextP, terminalName));

    entryP = (terminalListEntry_t *)malloc(sizeof(terminalListEntry_t));
    if (!entryP)
	{
		/* gen error message */
        TRACE_RETURN_TYPE(RSN_007, terminalListEntry_t *, entryP);
	}
	TRACE_ALLOC("plugin.c-33", entryP);

	/* initialise the terminal entry */

    /* Store the terminal name */
    UCcpy(entryP->name, terminalName);

    /* Create terminal and save its handle */
    entryP->handle = cniCreateOutputTerminal(&rc, 
		                                     contextP->nodeObject, 
									 		 terminalName);

	/* add this entry to the head of the terminal entry chain */
	entryP->next = contextP->outputTerminalListHead;
	contextP->outputTerminalListHead = entryP;

    TRACE_RETURN_TYPE(RSN_008, terminalListEntry_t *, entryP);
}

/******************************************************************************/
/* Deletes the storage for all terminal entries linked to the current context.*/
/******************************************************************************/
void deleteAllTerminalListEntries(context_t *contextP)
{
    TRACE_ENTRY("deleteAllTerminalListEntries", ("contextP=%.8x", contextP));

	/* delete input terminal entries */
	if (contextP->inputTerminalListHead)
	{
		TRACE_MID(TRACE_INFO, ("Deleting input terminals"));
		deleteTerminalListEntries(contextP->inputTerminalListHead);
	}

	/* delete output terminal entries */
	if (contextP->outputTerminalListHead)
	{
		TRACE_MID(TRACE_INFO, ("Deleting output terminals"));
		deleteTerminalListEntries(contextP->outputTerminalListHead);
	}

    TRACE_RETURN_VOID(RSN_009);
}

/******************************************************************************/
/* Deletes the storage for terminal entries linked to the supplied anchor.    */
/******************************************************************************/
void deleteTerminalListEntries(terminalListEntry_t *anchorP)
{
    terminalListEntry_t *entryP,
		                *nextEntryP;
    TRACE_ENTRY("deleteTerminalListEntries", ("anchorP=%.8x", anchorP));

	entryP = anchorP;

	while(entryP)
	{
		TRACE_MID(TRACE_INFO, ("processing entry '%S' at %.8x",
			                   entryP->name,
							   entryP));
		nextEntryP = entryP->next;

		/* no need to free the handle */

		/* free the entry itself */
		free(entryP);
		TRACE_FREE("plugin.c-33", entryP);

		entryP = nextEntryP;
	}

    TRACE_RETURN_VOID(RSN_010);
}

/******************************************************************************/
/* getInputTerminalHandle                                                     */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named input terminal.                           */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the input terminal list looking for the named input terminal.    */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/
void *getInputTerminalHandle(context_t *contextP,
							CciChar *terminalName)
{
    terminalListEntry_t *p;
    TRACE_ENTRY("getInputTerminalHandle", ("contextP=%.8x, terminalName=%S",
											contextP, terminalName));
	
	p = ((context_t *)contextP)->inputTerminalListHead;

    /* Having set the pointer to the start of the terminal list, traverse down
     * and compare each element in the list until the named terminal is found
     * then return the handle to that terminal object, otherwise return null   
     */
    while(p)
	{
        if (!UCcmp(p->name, terminalName))
            TRACE_RETURN_TYPE(RSN_011, void *, p->handle);
        p = p->next;
	}

    /* No match was found */
    TRACE_RETURN_TYPE(RSN_012, void *, NULL);
}

/******************************************************************************/
/* getOutputTerminalHandle                                                    */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context      (input) - pointer to the current context                     */
/*  terminalName (input) - name of terminal for handle to be returned         */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  The terminal object for a named output terminal.                          */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Traverses the output terminal list looking for the named output terminal.  */
/* When it finds the correct terminal, it returns the handle to it.           */
/*                                                                            */
/******************************************************************************/
void *getOutputTerminalHandle(context_t *contextP,
							  CciChar *terminalName)
{
    terminalListEntry_t *p;
    TRACE_ENTRY("getOutputTerminalHandle", ("contextP=%.8x, terminalName=%S",
											 contextP, terminalName));
	
	p = ((context_t *)contextP)->outputTerminalListHead;

    /* Having set the pointer to the start of the terminal list, traverse down
     * and compare each element in the list until the named terminal is found  
     * then return the handle to that terminal object, otherwise return null
     */
    while(p)
	{
        if (!UCcmp(p->name, terminalName))
            TRACE_RETURN_TYPE(RSN_013, void *, p->handle);
        p = p->next;
	}

    /* No match was found */
    TRACE_RETURN_TYPE(RSN_014, void *, NULL);
}

/******************************************************************************/
/* getAttribute                                                               */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context       (input)  - pointer to the current context                   */
/*  attributeName (input)  - attribute name for which value is to be          */
/*                           retrieved                                        */
/*  buffer        (output) - address of buffer into which attribute value is  */
/*                           copied                                           */
/*  bufferSize    (input)  - length of buffer                                 */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Gets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a report request is received that causes a retrieval    */
/* of the value of a node attribute. The broker will have verified that the   */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Return a character representation of the attribute value.               */
/* 2. Return a null string is data is sensitive and should not be displayed.  */
/* 3. Throw an exception if errors occur.                                     */
/*                                                                            */
/******************************************************************************/
int getAttribute(void *contextP,
				 CciChar *attributeName,
				 CciChar *buffer,
				 int bufsize)      
{
    attributeTableEntry_t *p;
	
    TRACE_ENTRY("getAttribute", ("contextP=%.8x, attributeName=%S, buffer=%.8x, "
		                         "bufsize=%d", 
								 contextP, attributeName, buffer, bufsize));

	p = ((context_t *)contextP)->attributeListHead;

    while (p != 0)
	{
        if (!UCcmp(p->name, attributeName))
		{
            TRACE_MID(TRACE_INFO, ("Found matching entry at %.8x", p));

            /* Bad return if value not set */
		    if (!p->attributeSet)
			{
				/* Unknown attribute */
                TRACE_MID(TRACE_INFO, ("Value is not set"));
				TRACE_RETURN_TYPE(RSN_015, int, CCI_EXCEPTION_UNKNOWN);
			}

            switch(p->type)
			{
				case CCI_VALUE_TYPE_BYTEARRAY:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_BYTEARRAY"));
					if (bufsize < (int)UClen((CciChar *)p->valuePtr))
					{
						/* deal with error */
                        TRACE_RETURN_TYPE(RSN_016, int, CCI_BUFFER_TOO_SMALL);
					}
					UCcpy(buffer, (CciChar *)p->valuePtr);
					break;

                case CCI_VALUE_TYPE_INTEGER:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_INTEGER"));
					if (bufsize < (int)UClen((CciChar *)p->valuePtr))
					{
						/* deal with error */
                        TRACE_RETURN_TYPE(RSN_016, int, CCI_BUFFER_TOO_SMALL);
					}
					UCcpy(buffer, (CciChar *)p->valuePtr);
                    break;

                case CCI_VALUE_TYPE_BOOL:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_BOOL"));
					if (*((int *)(p->valuePtr)))
					{
						UCcpy(buffer, P.TRUE_C);
					}
					else
					{
						UCcpy(buffer, P.FALSE_C);
					}
                    break;

                default:
                    TRACE_MID(TRACE_ERROR, ("Type not recognised"));
                    break;
			}

            /* Normal return */
            TRACE_MID(TRACE_INFO, ("Value is '%S'", buffer));
            TRACE_RETURN_TYPE(RSN_017, int, CCI_SUCCESS);
		}
        p = p->nextP;
	}

    /* Unknown attribute */
    TRACE_RETURN_TYPE(RSN_018, int, CCI_EXCEPTION_UNKNOWN);
}

/******************************************************************************/
/* getIntegerAttribute                                                        */
/******************************************************************************/
int getIntegerAttribute(IN context_t *contextP,
			   	        IN CciChar *attributeName,
						OUT int *valueP)
{
	CciChar buffer[200];
    TRACE_ENTRY("getIntegerAttribute", ("contextP=%.8x, attributeName=%S",
		                                contextP, attributeName));

	if (getAttribute((void *)contextP,
		             attributeName,
					 buffer,
					 sizeof(buffer)) != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("attribute '%S' not found", attributeName));
        TRACE_RETURN(RSN_019, TT_FAILED);
	}

	/* convert returned wide character string to integer */
	*valueP = UCToInt(buffer);
    TRACE_MID(TRACE_INFO, ("Returning value %d", *valueP));

    TRACE_RETURN(RSN_020, TT_SUCCESS);
}

/******************************************************************************/
/* getBooleanAttribute                                                        */
/******************************************************************************/
int getBooleanAttribute(IN context_t *contextP,
			   	        IN CciChar *attributeName,
						OUT bool_t *valueP)
{
	CciChar buffer[200];
    TRACE_ENTRY("getBooleanAttribute", ("contextP=%.8x, attributeName=%S",
		                                contextP, attributeName));

	if (getAttribute((void *)contextP,
		             attributeName,
					 buffer,
					 sizeof(buffer)) != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("attribute '%S' not found", attributeName));
        TRACE_RETURN(RSN_021, TT_FAILED);
	}

	/* convert returned wide character string to integer */
	if (!UCcmp(buffer, P.TRUE_C))
	{
	    *valueP = TRUE;
	}
	else
	{
	    *valueP = FALSE;
	}
    TRACE_MID(TRACE_INFO, ("Returning value %s", 
		                    *valueP == TRUE ? "TRUE" : "FALSE"));

    TRACE_RETURN(RSN_022, TT_SUCCESS);
}

/******************************************************************************/
/* getStringAttribute                                                         */
/******************************************************************************/
int getStringAttribute(IN context_t *contextP,
			           IN CciChar *attributeName,
					   INOUT int *bufferLengthP,
					   OUT CciChar *bufferP)
{

    TRACE_ENTRY("getStringAttribute", ("contextP=%.8x, attributeName=%S, *bufferLengthP=%d",
		                                contextP, attributeName, *bufferLengthP));
	if (getAttribute((void *)contextP,
		             attributeName,
					 bufferP,
					 *bufferLengthP) != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("attribute '%S' not found", attributeName));
		*bufferP = '\0';
		*bufferLengthP = 0;
        TRACE_RETURN(RSN_023, TT_FAILED);
	}
	*bufferLengthP = UClen(bufferP);
    TRACE_MID(TRACE_INFO, ("Returning string '%S'", bufferP));

    TRACE_RETURN(RSN_024, TT_SUCCESS);
}

/******************************************************************************/
/* getHexAttribute                                                            */
/******************************************************************************/
int getHexAttribute(IN context_t *contextP,
		            IN CciChar *attributeName,
				    INOUT int *bufferLengthP,
				    OUT CciChar *bufferP)
{
	CciChar buffer[2000];
	CciChar textString[MAX_STRING_SIZE + 1];
	int stringLength;

    TRACE_ENTRY("getHexAttribute", ("contextP=%.8x, attributeName=%S, *bufferLengthP=%d",
		                             contextP, attributeName, *bufferLengthP));

	memset(buffer, 0, sizeof(buffer));
	if (getAttribute((void *)contextP,
		             attributeName,
					 buffer,
					 sizeof(buffer)) != CCI_SUCCESS)
	{
        TRACE_MID(TRACE_ERROR, ("attribute '%S' not found", attributeName));
		UCcpy(bufferP, P.NULL_STRING);
		*bufferLengthP = 0;
        TRACE_RETURN(RSN_025, TT_FAILED);
	}

	if (!UClen(buffer))
	{
		UCcpy(bufferP, P.NULL_STRING);
		*bufferLengthP = 0;
        TRACE_RETURN(RSN_026, TT_SUCCESS);
	}

	stringLength = sizeof(textString);
	/* now convert from hex to char array, assume no embedded nulls */
	if (convertNameFromHex(buffer,
		                   &stringLength,
		                   textString) != TT_SUCCESS)
	{
	    TRACE_RETURN(RSN_027, TT_FAILED);
	}
	*bufferLengthP = stringLength;
	UCcpy(bufferP, textString);
    TRACE_MID(TRACE_INFO, ("returning string length %d", stringLength));

    TRACE_RETURN(RSN_028, TT_SUCCESS);
}

/******************************************************************************/
/* commonSetAttribute                                                         */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context        (input)  - pointer to the current context                  */
/*  attributeName  (input)  - attribute name for which value is to be set     */
/*  attributeValue (input) -  value to be set for the attribute               */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Sets the value of an attribute on a specific node instance. It is invoked  */
/* by the broker when a configuration request is received that attempts to    */
/* set the value of a node attribute. The broker will have verified that the  */
/* attribute name is valid for the node.                                      */
/*                                                                            */
/* This function has 3 responsibilities:                                      */
/*                                                                            */
/* 1. Verify value of attribute is correctly specified, otherwise a           */
/*    configuration exception should be thrown.                               */
/* 2. Store the attribute value in the context.                               */
/* 3. Throw a configuration exception if errors occur.                        */
/*                                                                            */
/******************************************************************************/
int commonSetAttribute(CciContext *contextP,
				       CciChar *attributeName,
				       CciChar *attributeValue)
{
    attributeTableEntry_t *p;
	
    TRACE_ENTRY("commonSetAttribute", ("contextP=%.8x, attributeName=%S, "
		                               "attributeValue='%S'",
								       contextP, attributeName, attributeValue));

	p = ((context_t *)contextP)->attributeListHead;

    while (p)
	{
		if (!UCcmp(p->name, attributeName))
		{
            TRACE_MID(TRACE_INFO, ("Found matching entry at %.8x", p));
			switch(p->type)
			{
				case CCI_VALUE_TYPE_BYTEARRAY:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_BYTEARRAY"));

					UCcpy((CciChar *)(p->valuePtr), attributeValue);

					/* we also store the label name in node context */
					if (!UCcmp(attributeName, P.ATTR_LABEL))
					{
                        ((context_t *)contextP)->nodeLabel = UCdup(attributeValue);
						TRACE_ALLOC("plugin.c-34", ((context_t *)contextP)->nodeLabel);
					}
					break;

				case CCI_VALUE_TYPE_INTEGER:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_INTEGER"));
					UCcpy((CciChar *)(p->valuePtr), attributeValue);
					break;

				case CCI_VALUE_TYPE_BOOL:
                    TRACE_MID(TRACE_INFO, ("Type is CCI_VALUE_TYPE_BOOL"));
					if (!UCcmp(attributeValue, P.FALSE_C))
					{
                        TRACE_MID(TRACE_INFO, ("Value is 'false'"));
						*((int *)(p->valuePtr)) = 0;
					}
					else
					{
                        TRACE_MID(TRACE_INFO, ("Value is 'true'"));
						*((int *)(p->valuePtr)) = 1;
					}
					break;

				default:
					break;
			}

			/* remember that attribute is now set */
			p->attributeSet = TRUE;

			/* Normal return */
            TRACE_RETURN(RSN_029, TT_SUCCESS);
		}
        p = p->nextP;
	}

    /* Error...attribute not found */
    TRACE_RETURN(RSN_030, CCI_EXCEPTION_UNKNOWN);
}

/******************************************************************************/
/* getAttributeName                                                           */
/*                                                                            */
/* Parameters:                                                                */
/*                                                                            */
/*  context    (input)  - pointer to the current context                      */
/*  index      (input)  - index of the attribute for which the name is to be  */
/*                        retrieved                                           */
/*  buffer     (output) - address of buffer into which attribute name is      */
/*                        copied                                              */
/*  bufferSize (input)  - length of buffer                                    */
/*                                                                            */
/* Returns:                                                                   */
/*                                                                            */
/*  0 - success                                                               */
/*  1 - failure, unknown attribute name                                       */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* Returns the name of an attribute by a given index. It is invoked by the    */
/* broker when it requires the names of attributes supported by a particular  */
/* instance of a node. The function must guarantee to return the attribute     */
/* in a known, defined order and to return the attribute name represented by  */
/* the index parameter.                                                       */
/*                                                                            */
/******************************************************************************/
int getAttributeName(void* contextP,
					 int index,
					 CciChar* buffer,
					 int bufferSize)
{
    int count = 0;
    attributeTableEntry_t *p;
    TRACE_ENTRY("getAttributeName", ("contextP=%.8x, index=%d, "
		                         "buffer=%.8x, bufferSize=%d",
								 contextP, index, buffer, bufferSize));
	
	p = ((context_t *)contextP)->attributeListHead;

    /* having set the pointer to the start of the attribute list, traverse down*/
    /* and compare each element in the list until the indexed element is found */
    /* then return the name of the attribute in the supplied buffer.           */
	/* However, we only want to return attribute names where the value has been*/
	/* set.  Therefore, we exclude any unset attributes in our indexing scheme.*/

    while (p)
	{
		if (!p->attributeSet || !p->customAttribute)
		{
            TRACE_MID(TRACE_INFO, ("Ignoring unset or non-custom attribute "
				                   "name '%S'", p->name));
		    p = p->nextP;
			continue;
		}

		if (count == index)
		{
            TRACE_MID(TRACE_INFO, ("Returning name '%S'", p->name));
			UCcpy(buffer, p->name);
            TRACE_RETURN_TYPE(RSN_031, int, CCI_SUCCESS);
		}
		count++;
		p = p->nextP;
	}

    /* Invalid index */
    TRACE_RETURN_TYPE(RSN_032, int, CCI_EXCEPTION_UNKNOWN);
}

/******************************************************************************/
/* convertNameFromHex                                                         */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
int convertNameFromHex(IN CciChar *inHexString,
					   INOUT int *bufferLengthP,
					   OUT CciChar *buffer)
{
	const char validHexValues[] = "0123456789abcdefABCDEF";
	char stringName[MAX_STRING_SIZE+1],
	     mbString[MAX_STRING_SIZE + 1];
	int inputLength,
		i,
		j;
	char hexString[3],
		 char4[4];
	int *intP = (int *)char4;

    TRACE_ENTRY("convertNameFromHex", ("inHexString=%S, bufferLength=%d, "
		                               "buffer=%.8x",
		                               inHexString, *bufferLengthP, buffer));

	/* Not sure how to convert 'hexified' unicode string into standard unicode string,
	 * so we'll first convert the incoming string from unicode to multibyte, do the 
	 * conversion and then convert the result back into unicode.
	 */
    UCToStr(inHexString, 
		    mbString, 
		    sizeof(mbString),    
			&inputLength);

    TRACE_MID(TRACE_INFO, ("Length of input string is %d", inputLength));

	if (inputLength > sizeof(stringName) - 1)
	{
        TRACE_MID(TRACE_ERROR, ("Input string is too long"));
	    TRACE_RETURN(RSN_033, TT_FAILED);
	}

    /* ensure that all characters are valid hex chars */
	for (i=0; i<inputLength; i++)
	{
		if (!strchr(validHexValues, mbString[i]))
		{
			/* invalid hex string */
	        TRACE_RETURN(RSN_034, TT_FAILED);
		}
	}
    
	TRACE_MID(TRACE_INFO, ("Converting hex characters"));
	hexString[2] = '\0';
	
	for (j=0,i=0; i<inputLength; )
	{
        hexString[0] = mbString[i++];
		hexString[1] = mbString[i++];

		/* convert string to upper case */
		strToUpper(hexString);
		TRACE_MID(TRACE_INFO, ("Processing hexString '%s'",
			                   hexString));
        
		/* convert from hex to char */
		char4[0] = ' ';
		sscanf(hexString, "%X", intP);
		TRACE_MID(TRACE_INFO, ("Processed hexString '%s' to '%c'",
			                   hexString, char4[0]));
		/* add to converted name */
		stringName[j++] = char4[0];
	}
    
	stringName[j] = '\0';
	TRACE_MID(TRACE_INFO, ("Processed stringName is '%s'",
		                    stringName));
	TRACE_MID(TRACE_INFO, ("Length of converted string is %d", j));

	/* convert multi byte string back to unicode */
	strToUC(stringName,
		    *bufferLengthP,
            buffer,
			NULL);

	/* 
	 * Return length of converted string to caller - can't use string functions to 
	 * determine the length since the string may contain embedded null characters.
	 */
	*bufferLengthP = j;

	TRACE_RETURN(RSN_035, TT_SUCCESS);
}

